# Example for mac/mac-tdmadama.{cc,h} module
# CONFIGURATION VARIABLES ###########################################

;# testing == 1  enables tracing for debugging purposes

set testing           1
set myseed            1 
set duration       5000.0
set s_int             1.0

# Satellite Dependent

set frame_duration    0.04
set delayTER          50ms
set delayRCST         10ms

set capTER           100Mb
set capSAT           515200 

set k_frame_          1.0 
set rcst_buffer_    500 
set hub_buffer_   500 

set slotsize_         184 
set state_period       0.0417
set superframe_      25

##### Radio Resource Management #############

Mac/TdmaDama set fixed_frame_ 0

Allocator/FODA set ignore_loss_ 1 
Allocator/FODA set delay_ 0.797
Allocator/FODA set def_slots_ 0
Allocator/FODA set xpercent_ 1.0

Requester/RVBDC set alpha_ 1.0
Requester/RVBDC set k_ 1.0
Requester/RVBDC set min_rvbdc_ 1.0
Requester/RVBDC set max_rvbdc_ 0

Queue/DropTail set drop_front_ true

############## Traffic Flows #############

# Data Flows
set data(dir)       upload
set data(next)          3.0
set data(bunch)          5
set data(packet)      1500

Agent/TCP set window_ 1000
Agent/TCP set windowInit_ 2.0
Agent/TCP set minrto_ 1.0 


##########  loading alternative configurations ###########
set cmdline [lindex $argv 0]
if { $cmdline != "scpc" && $cmdline !="bod" } {
	puts "usage: ./test1.tcl \[bod|scpc\]"
	exit 1
}

ns-random $myseed
expr srand($myseed)

##########################################################

set slotframe_   [expr round(($capSAT*$frame_duration)/(8*$slotsize_))]
set data(load)   [expr $data(bunch)*$data(packet)*8/($capSAT*$data(next))]
puts "data(load) [format %.4f $data(load)]"

if { $cmdline == "scpc" } {
	Mac/TdmaDama set fixed_frame_ 1
	Mac/TdmaDama set slots1_ $slotframe_
	Mac/TdmaDama set offset_ -1
}

# scheduling data flows
set data(index) 0
proc new-data {} { 
	global ns wl hq data

	set data_app [new Application/FTP]
	set data_tcp [new Agent/TCP/Sack1]
	set data_snk [new Agent/TCPSink/Sack1]
	$data_tcp set packetSize_ $data(packet)
	$data_tcp set time_start [$ns now] 

	$ns attach-agent $wl $data_tcp 
	$ns attach-agent $hq $data_snk
	$ns connect $data_tcp $data_snk
	$data_app attach-agent $data_tcp
	set exp [expr -log(rand())]
	$data_app produce [expr $exp*$data(bunch)]
	
	set exp [expr -log(rand())]
	set next [expr [$ns now] + $exp*$data(next)]
	incr data(index)
	$ns at $next "new-data"
}


Mac/TdmaDama set max_slot_num_ $slotframe_
Mac/TdmaDama set slot_packet_len_ $slotsize_
Mac/TdmaDama set bandwidth_ $capSAT
Mac/TdmaDama set num_frame_ $superframe_
Mac/TdmaDama set aligned_ 1

Allocator/FODA set k_frame_ $k_frame_
Allocator/FODA set maxg_ $slotframe_
Allocator/FODA set ming_ 1 

# Creating scenario  ##########################

set ns [new Simulator]
set stats_con [open stats.tr w]
if { $testing == 1 } {
	# Tracing enabling must 
	# precede link and node creation 
	set f0 [open out.tr w]
	$ns trace-all $f0
}

# nodes
set hq    [$ns node]
set wl    [$ns node]

set opt(chan)           Channel/Sat
set opt(bw_up)		$capSAT        ;# uplink bandwidth
set opt(bw_down)	$capSAT        ;# downlink bandwidth
set opt(phy)            Phy/Sat
set opt(mac)            Mac/TdmaDama
set opt(ifq)            Queue/DropTail
set opt(qlim)		230            ;# queue size (pkts)
set opt(ll)             LL/Sat
set opt(wiredRouting)   ON

# Configure bent-pipe satellite
$ns node-config -satNodeType geo-repeater \
       	        -llType $opt(ll) \
       	        -ifqType $opt(ifq) \
       	        -ifqLen $opt(qlim) \
       	       	-macType $opt(mac) \
		-phyType Phy/Repeater \
		-channelType $opt(chan) \
		-downlinkBW $opt(bw_down)  \
		-wiredRouting $opt(wiredRouting)

# GEO satellite at 13 degrees longitude East (Hotbird 6)
set sat [$ns node]

$sat set-position 13
set lat 43.71
set lon 10.38

$ns node-config -satNodeType terminal \
       	        -llType $opt(ll) \
       	        -ifqType $opt(ifq) \
       	        -ifqLen $opt(qlim) \
       	       	-macType $opt(mac) \
       	       	-phyType $opt(phy) \
       	       	-channelType $opt(chan) \
       	       	-downlinkBW $opt(bw_down) \
       	       	-wiredRouting $opt(wiredRouting)

set hub   [$ns node]
set rcst  [$ns node]

$hub set-position $lat $lon
$rcst set-position $lat $lon

$hub add-gsl geo $opt(ll) $opt(ifq) $hub_buffer_ $opt(mac) \
	$opt(bw_up) $opt(phy) [$sat set downlink_] [$sat set uplink_]

$rcst add-gsl geo $opt(ll) $opt(ifq) $rcst_buffer_ $opt(mac) \
	$opt(bw_up) $opt(phy) [$sat set downlink_] [$sat set uplink_]

set gat_mac [$hub set mac_(0)]
set ter_mac [$rcst set mac_(0)]
$gat_mac set hub_ 1

if { $testing == 1 } {
	set ev_file [open event.tr w]
	$hub trace-event $ev_file
	$rcst trace-event $ev_file
	$ns trace-all-satlinks $f0
}


# create node requester object
set streq [$hub install-requester Requester/Constant]
set dtreq [$rcst install-requester Requester/RVBDC]

# Network Control Center
set ncc [$rcst install-allocator Allocator/FODA]

# We use centralized routing
set satrouteobject_ [new SatRouteObject]
$satrouteobject_ compute_routes

$ns duplex-link $hq $hub $capTER $delayTER DropTail
$ns duplex-link $wl $rcst $capTER $delayRCST DropTail


set start 1.0

### Drawing diagrams

Agent/TCP/Sack1 instproc done {} {
	global ns stats_con
	set init [$self set time_start]
	set dur [expr [$ns now] - $init]
	puts $stats_con [format "%10.2f %8.4f" $init $dur]

}

set function_ci {
	function A = autocorr(x)
		A = x - mean(x);
		A = fftconv(A, rot90(A,2));
		A = A(length(x) : 2*length(x)-1);
		A /= length(x);
		A /= A(1);
	end
	
	function t = corrlen(x)
	
	 if (rows(x)==1)
	   x = x';
	 endif
	 [l,n] = size(x);
	 weights = 1 - [0.5, [1:l-1]/(l-1)];
	 clen = zeros(1,n);
	 for i=1:n
	   clen(i) = weights * autocorr(x(:,i));
	 endfor
	 t = max(1,clen);
	
	end
	
	input = load("stats.tr");
	input = input(:,2);
	
	step = ceil(corrlen(input));
	data = input(1:step:length(input));
	
	confidence = 0.95;
	
	m = mean(data);
	s = std(data);
	p = abs(tinv( (1-confidence)/2 , length(data) - 1));
	p = p * s / sqrt(length(data));
	
	ci = [m ; p];
	
	h = reshape(ci, 1, size(ci)(1) * size(ci)(2));
	
	printf("%10.5g %6.4f",h);
	printf("\n");
}

set plot_res1 {
	set xlabel "time [s]"
	set ylabel "bandwidth [b/s]"
	
	set xrange [0:500]
	set yrange [0:600000]
	
	set term postscript color solid eps 
	set output "result1.eps"
	
	plot "q_up.dat" u 1:4 t 'used cap.' w l 3, \
	     "curall.tr" u (floor($2)):($5*515200/14) smooth unique \
	     t 'alloc cap.' w l 1
}

proc finish-sim {} {
	global ns rcst hub stats_con cmdline
	global function_ci plot_res1

	$ns flush-trace
	$ns halt

	close $stats_con

	exec grep " [$rcst id] [$hub id] " out.tr | \
	cut -d " " -f 1,2,3,4,5,6,7,8,9,10,11,12 > q_up

	exec analy q_up 1.0 
	exec grep curall event.tr > curall.tr
	exec grep reques event.tr > reques.tr

	puts "mean trans.time  [exec echo $function_ci | octave -q]"

	exec echo $plot_res1 | gnuplot
	exec mv result1.eps result1-${cmdline}.eps
	exec mv stats.tr stats-test1-${cmdline}.tr
	exec mv q_up.dat q_up-test1-${cmdline}.dat
	exec mv out.tr out-test1-${cmdline}.tr 
	exec mv event.tr event-test1-${cmdline}.tr
	exec rm q_up reques.tr curall.tr
	
}

$ns at $start "new-data"
$ns at $duration "finish-sim"

$ns run 

