/*
 * mac-tdmadama.cc
 * Copyright (C) 1999 by the University of Southern California
 * $Id: mac-tdma.cc,v 1.15 2005/09/18 23:33:33 tomh Exp $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 * The copyright of this module includes the following
 * linking-with-specific-other-licenses addition:
 *
 * In addition, as a special exception, the copyright holders of
 * this module give you permission to combine (via static or
 * dynamic linking) this module with free software programs or
 * libraries that are released under the GNU LGPL and with code
 * included in the standard release of ns-2 under the Apache 2.0
 * license or under otherwise-compatible licenses with advertising
 * requirements (or modified versions of such code, with unchanged
 * license).  You may copy and distribute such a system following the
 * terms of the GNU GPL for this module and the licenses of the
 * other code concerned, provided that you include the source code of
 * that other code when and as the GNU GPL requires distribution of
 * source code.
 *
 * Note that people who make modified versions of this module
 * are not obligated to grant this special exception for their
 * modified versions; it is their choice whether to do so.  The GNU
 * General Public License gives permission to release a modified
 * version without this exception; this exception also makes it
 * possible to release a modified version which carries forward this
 * exception.
 *
 */

// by Raffaello Secchi (raffaello.secchi@isti.cnr.it), CNR/ISTI

// Preamble TDMA MAC layer for single hop.
// Centralized slot assignment computing.

#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"

#define DEBUG_MAC 1

//#include <debug.h>

#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "mac-tdmadama.h"
#include "wireless-phy.h"
#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"
#include "arp.h"
#include "ll.h"
#include "mac.h"

#define SET_RX_STATE(x)			\
{					\
	rx_state_ = (x);			\
}

#define SET_TX_STATE(x)				\
{						\
	tx_state_ = (x);				\
}

double MacTdmaDama::slot_time_ = 0;
double MacTdmaDama::start_time_ = 0;
int MacTdmaDama::active_node_ = 0;
int MacTdmaDama::max_slot_num_ = 0;
int MacTdmaDama::num_requests_ = 0;
int *MacTdmaDama::tdma_schedule_nxt_ = NULL;
double *MacTdmaDama::requestv_ = NULL;
TdmaAllocator *MacTdmaDama::allocator_ = NULL;
int  MacTdmaDama::num_frame_ = 1;

int MacTdmaDama::tdma_ps_ = 0;
int MacTdmaDama::tdma_pr_ = 0;
int hdr_tdmadama::offset_;

/* Phy specs from 802.11 */
static PHY_MIB PMIB = {
	DSSS_CWMin, DSSS_CWMax, DSSS_SlotTime, DSSS_CCATime,
	DSSS_RxTxTurnaroundTime, DSSS_SIFSTime, DSSS_PreambleLength,
	DSSS_PLCPHeaderLength
};	

/* Timers */
void MacTdmaDamaTimer::start(Packet *p, double time)
{
	Scheduler &s = Scheduler::instance();
	assert(busy_ == 0);
  
	busy_ = 1;
	paused_ = 0;
	stime = s.clock();
	rtime = time;
	assert(rtime >= 0.0);
  
	s.schedule(this, p, rtime);
}

void MacTdmaDamaTimer::stop(Packet *p) 
{
	Scheduler &s = Scheduler::instance();
	assert(busy_);
  
	if(paused_ == 0)
		s.cancel((Event *)p);

	// Should free the packet p.
	Packet::free(p);
  
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
}

/* Slot timer for TDMA scheduling. */
void SlotTdmaDamaTimer::handle(Event *e)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
  
	mac->slotHandler(e);
}

/* Receive Timer */
void RxPktTdmaDamaTimer::handle(Event *e) 
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
  
	mac->recvHandler(e);
}

/* Send Timer */
void TxPktTdmaDamaTimer::handle(Event *e) 
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
	
	mac->sendHandler(e);
}

static class MacTdmaDamaClass : public TclClass {
public:
	MacTdmaDamaClass() : TclClass("Mac/TdmaDama") {}
	TclObject* create(int, const char*const*) {
		return (new MacTdmaDama(&PMIB));
	}
} class_mac_tdma;

static class TdmaDamaHeaderClass : public PacketHeaderClass {
public:
	TdmaDamaHeaderClass(): PacketHeaderClass("PacketHeader/TdmaDama",
				    sizeof(hdr_tdmadama)) {
		bind_offset(&hdr_tdmadama::offset_);
	}
} class_hdr_tdmadama;


// Mac Tdma definitions
// Frame format: Slot1 Slot2 Slot3...
MacTdmaDama::MacTdmaDama(PHY_MIB* p) : 
	Mac(), requester_(0), et_(0), dump_alloc_(0), 
	mhSlot_(this), mhTxPkt_(this),  mhRxPkt_(this) {

	/* Global variables setting. */
	// Setup the phy specs.
	static int slot_pointer = 0;

	phymib_ = p;

	// bind parameters of MacTdmaDama object
	bind("slot_packet_len_", &slot_packet_len_);
	bind("max_slot_num_", &max_slot_num_);
	bind("num_frame_", &num_frame_);
	bind("dump_alloc_", &dump_alloc_);
	bind("abw_",&abw_);	
	bind("util_weight_",&util_weight_);
	bind("enableQS_",&enableQS_);
	bind("delayQS_",&delayQS_);
	bind("algorithm_",&algorithm_);	
	bind("aligned_",&aligned_);
	bind("fixed_frame_",&fixed_frame_);
	bind("slots1_",&slots1_);
	bind("m_frame_",&m_frame_);
	bind("offset_",&offset_);
	bind("hub_",&hub_);
	bind("fragsize_",&fragsize_);

	/* statistics */
	bind("corrupted_slots_",&corrupted_slots_);
	bind("used_slots_",&used_slots_);
	bind("successful_slots_",&successful_slots_);
	bind("total_slots_",&total_slots_);
	bind("slot_num_", &slot_num_);

	flag_used_slot_ = 0;	

	// Calculate the slot time based on the MAX allowed data length.
	slot_time_ = DATA_Time(slot_packet_len_);	

	// Initialize the tdma schedule data structures.
	if ( !tdma_schedule_nxt_ ) {
		tdma_schedule_nxt_ = new int[max_slot_num_];
		requestv_ = new double[max_slot_num_];

		for(int i=0; i<max_slot_num_; i++) {
			tdma_schedule_nxt_[i] = SLOT_UNALLOC;
			requestv_[i] = SLOT_NOREQ;
		}
	}

	for(int i=0; i<MAC_MAX_TERM; i++)
		tdma_schedule_[i] = SLOT_UNALLOC;

	/*
	if ( active_node_ < max_slot_num_) {
		tdma_schedule_[active_node_] = active_node_;
		tdma_schedule_nxt_[active_node_] = active_node_;
	}
	*/

	// Record the initial active node number.
	active_node_++;
	
	/* Do each node's initialization. */
	
	// Initial channel / transceiver states. 
	tx_state_ = rx_state_ = MAC_IDLE;

	// Initialy, the radio is off. NOTE: can't use radioSwitch(OFF) here.
	radio_active_ = 0;

	// Record the start time of the new schedule.
	start_time_ = NOW;
	slot_start_time_ = NOW;

	/* Seperate slot_num_ and the node id: 
	   we may have flexibility as node number changes.
	*/
	slot_num_ = slot_pointer++;
	
	frame_ = 0;

	if ( allocator_ )
		allocator_->init();
		
	/* Deal with preamble. */
	// Can't send anything in the first frame.
	next_slot_ = FIRST_ROUND;
	
	residue_ = 0;
	
	//Start the Slot timer...
	mhSlot_.start((Packet *) (& intr_), 0);

	// available bandwidth estimation
	abw_ = 0;
	int_begin = 0;

        /* no errors by default */
	em_ = NULL;
	btp_ = NULL;

	/* loss model */
	packet_lost_ = 0;
	slot_lost_ = 0;

	m_Fcount = 0;

}

/* similar to 802.11, no cached node lookup. */
int MacTdmaDama::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "mac-queue") == 0) {
			mac_queue_ = (Queue*) TclObject::lookup(argv[2]);
			if(mac_queue_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		}
		if (strcmp(argv[1], "log-target") == 0) {
			logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			if(logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		}
		if (strcmp(argv[1], "eventtrace") == 0) {
			et_ = (EventTrace *)TclObject::lookup(argv[2]);
			if (!et_)
				return TCL_ERROR;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "errormac") == 0) {
			em_ = (ErrorModel *)TclObject::lookup(argv[2]);
			if (!em_)
				return TCL_ERROR;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "btp-error") == 0) {
			btp_ = (ErrorModel *)TclObject::lookup(argv[2]);
			if (!btp_)
				return TCL_ERROR;
			return (TCL_OK);
		}
	}
	return Mac::command(argc, argv);
}


/* ======================================================================
   Debugging & Tracing Routines
   ====================================================================== */
void MacTdmaDama::trace_pkt(Packet *p) 
{
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac_tdma* dh = HDR_MAC_TDMA(p);
	u_int16_t *t = (u_int16_t*) &dh->dh_fc;

	fprintf(stderr, "\t[ %2x %2x %2x %2x ] %x %s %d\n",
		*t, dh->dh_duration,
		ETHER_ADDR(dh->dh_da), ETHER_ADDR(dh->dh_sa),
		index_, packet_info.name(ch->ptype()), ch->size());
}

void MacTdmaDama::dump(char *fname)
{
	fprintf(stderr, "\n%s --- (INDEX: %d, time: %2.9f)\n", fname, 
		index_, Scheduler::instance().clock());
	
	fprintf(stderr, "\ttx_state_: %x, rx_state_: %x, idle: %d\n", 
		tx_state_, rx_state_, is_idle());
	fprintf(stderr, "\tpktTx_: %lx, pktRx_: %lx, callback: %lx\n", 
		(long) pktTx_, (long) pktRx_, (long) callback_);
}

void MacTdmaDama::watch_alloc(){
	
	for(int i=0; i<max_slot_num_; i++)
		if (tdma_schedule_[i] == SLOT_UNALLOC)
			printf("[N]");
		else
			printf("[%d]",tdma_schedule_[i]);
	printf("\n");	

}


void MacTdmaDama::trace_event(char *event, int val)
{
	
	if (et_ == NULL) return;
	char *wrk = et_->buffer();
	
	if (dump_alloc_)
	   if ( strcmp(event,"req") != 0 &&
		strcmp(event,"nxt") != 0 && 
		strcmp(event,"cur") != 0 )
		return;
		
	if (wrk != 0) {
	   if (strcmp(event,"cur") == 0) {
	      static int *xc = new int[active_node_];
	      for(int i=0; i<active_node_; i++)
		   xc[i] = 0;
	      sprintf(wrk, "<%d> %lf %sslt ", index_, NOW, event);
	      for(int i=0; i<max_slot_num_; i++)
		 if (tdma_schedule_nxt_[i] == SLOT_UNALLOC)
		    sprintf(wrk+strlen(wrk), "[N]");
		 else{
		    sprintf(wrk+strlen(wrk), "[%d]", tdma_schedule_nxt_[i]);
		    xc[tdma_schedule_nxt_[i]]++;
		 }
	      sprintf(wrk+strlen(wrk), "\n<%d> %lf %sall ", index_, NOW, event);
	      for(int i=0; i<active_node_; i++)
		   sprintf(wrk+strlen(wrk), "%d ", xc[i]);
	   } else if (strcmp(event,"reques") == 0) {
	      sprintf(wrk, "<%d> %lf %s ", index_, NOW, event);
	      for(int i=0; i<active_node_; i++ )
		 sprintf(wrk+strlen(wrk),"%1.3lf ",requestv_[i]);
	   } else if (strcmp(event,"nxt") == 0) {
	      static int *xs = new int[active_node_];
	      for(int i=0; i<active_node_; i++)
		   xs[i] = 0;
	      sprintf(wrk, "<%d> %lf %sslt ", index_, NOW, event);
	      for(int i=0; i<max_slot_num_; i++ )
	   	if (allocator_->schedule_nxt_[i] == SLOT_UNALLOC)
		   sprintf(wrk+strlen(wrk), "[N]");
		else{
		   sprintf(wrk+strlen(wrk), "[%d]", allocator_->schedule_nxt_[i]);
		   xs[allocator_->schedule_nxt_[i]]++;
		}
	      sprintf(wrk+strlen(wrk), "\n<%d> %lf %sall ", index_, NOW, event);
	      for(int i=0; i<active_node_; i++)
		   sprintf(wrk+strlen(wrk), "%d ", xs[i]);
	   } else if (strcmp(event,"req") == 0) {
		     static double *prev = new double[active_node_];
		     static int   *prevb = new int [active_node_];
		     static int flg = 0;
		     
		     if (!flg) {
     		     	for(int i=0; i<active_node_; i++){
				prev[i] = 0.0;
				prevb[i] = 0;
		     	}
		        flg = 1;
		     }

		     int queue;
    		     
		     double dt = 0;
		     double db = 0;
		     	     
		     dt = NOW - prev[index_];
		     
		     
		     db = 8*((int)requester_->qMon_->barrivals() - 
		          (int)requester_->qMon_->bdrops()-prevb[index_]);

		     prevb[index_] = (int)requester_->qMon_->barrivals() - 
		                     (int)requester_->qMon_->bdrops();
		
		     queue = (int)requester_->qMon_->barrivals() - 
			     (int)requester_->qMon_->bdepartures() -
		             (int)requester_->qMon_->bdrops();
		     

		     sprintf(wrk, "<%d> %lf reqpar %d %lf", index_, NOW,
		   	queue,db/dt);
		     
		     prev[index_] = NOW;
		     
	   } else 
	      sprintf(wrk, "<%d> %lf %s %d", index_, NOW, event, val);
	}

	et_->trace();
}


/* ======================================================================
   Packet Headers Routines
   ====================================================================== */

int MacTdmaDama::hdr_dst(char* hdr, int dst )
{

	struct hdr_mac_tdma *dh = (struct hdr_mac_tdma*) hdr;
	if(dst > -2)
		STORE4BYTE(&dst, (dh->dh_da));
	return ETHER_ADDR(dh->dh_da);
}

int MacTdmaDama::hdr_src(char* hdr, int src )
{

	struct hdr_mac_tdma *dh = (struct hdr_mac_tdma*) hdr;
	if(src > -2)
		STORE4BYTE(&src, (dh->dh_sa));
  
	return ETHER_ADDR(dh->dh_sa);
}

int MacTdmaDama::hdr_type(char* hdr, u_int16_t type) 
{
	struct hdr_mac_tdma *dh = (struct hdr_mac_tdma*) hdr;
	if(type)
		STORE2BYTE(&type,(dh->dh_body));
	return GET2BYTE(dh->dh_body);
}

/* Test if the channel is idle. */
int MacTdmaDama::is_idle() {
	if(rx_state_ != MAC_IDLE)
		return 0;
	if(tx_state_ != MAC_IDLE)
		return 0;
	return 1;
}

/* calculate frame used rate */
void MacTdmaDama::bandest(int psize) {
	double bws;    // bandwidth sample
	double int_end = NOW;
	double decay;

	switch(algorithm_) {
	    case 1:
		if (int_begin < int_end) {
			bws = (psize*8)/(int_end - int_begin);
			decay = exp(-util_weight_*(int_end - int_begin));
			abw_ = bws + (abw_ - bws) * decay;
			int_begin = int_end;
	   	}
		break;
	     
	    case 2:
	    	bws = 0;
		for(int i=0; i<max_slot_num_; i++) {
			if (tdma_schedule_[i]>=0) bws++;
		}
		abw_ = bandwidth_ * bws /max_slot_num_;
	}
}

double MacTdmaDama::utilization() {
	return abw_/bandwidth_;
}

double MacTdmaDama::peak_utilization() {
	// TO FIX !!
	return abw_/bandwidth_;
}

/* Do slot re-scheduling */
void MacTdmaDama::re_schedule() {

	/* make the requests */
	if (frame_ == 0 && requester_) {
		trace_event("req",slot_num_);
		requester_->dorequest();
	}

	frame_ = (frame_ + 1) % num_frame_;

	/* The frame structures is known a priori*/
	if ( fixed_frame_ ) {

		if ( slot_num_ == 0 )
			return;

		for(int i=0; i<max_slot_num_; i++)
		    tdma_schedule_[i] = 0;
		
		m_Fcount = (m_Fcount + 1) % m_frame_;
		if ( m_Fcount != 0)
			return;

		int ofs = offset_;
		if ( ofs < 0 ) 
			ofs = rand() % (max_slot_num_ - slots1_ + 1);

	        if ( btp_ && btp_->corrupt(NULL) ) {
	     	   for(int i=ofs; i<slots1_+ofs; i++)
		        tdma_schedule_[i] = SLOT_UNALLOC;
	        } else {
	   	   for(int i=ofs; i<slots1_+ofs; i++)
		        tdma_schedule_[i] = 1;
		}	   	
	
		if ( slot_num_ == 1 )
		        total_slots_ += slots1_;

		return;
	}

	// int count = total_slots_;
	for(int i=0; i<max_slot_num_; i++)
	   if ( tdma_schedule_nxt_[i] == slot_num_ )
	        total_slots_++;

	// if ( slot_num_ == 1 )
	//	printf("%lf %d\n",NOW, total_slots_ - count);

	/* BTP lost */
	if ( btp_ && btp_->corrupt(NULL) ) {
	   for(int i=0; i<max_slot_num_; i++)
		if ( tdma_schedule_nxt_[i] == slot_num_ )
	   	     tdma_schedule_[i] = SLOT_UNALLOC;
	   return;
	}  

	/* apply next schedule */
	for(int i=0; i<max_slot_num_; i++)
	     if ( tdma_schedule_nxt_[i] == slot_num_ )
		    tdma_schedule_[i] = tdma_schedule_nxt_[i];
		 else
			tdma_schedule_[i] = SLOT_UNALLOC;

}

/* Do slot re-allocation */
void MacTdmaDama::re_allocate() {

	num_requests_++;

	// printf("%lf %d %d\n", NOW, num_requests_, slot_num_);

	if (num_requests_ < active_node_)
		return;

	if (allocator_)
		allocator_->allocation();

	num_requests_ = 0;
}

/* To handle incoming packet. */
void MacTdmaDama::recv(Packet* p, Handler* h) {
	struct hdr_cmn *ch = HDR_CMN(p);

	u_int32_t dst;
	struct hdr_mac_tdma* dh;

	/* Incoming packets from phy layer, send UP to ll layer. 
	   Now, it is in receiving mode. 
	*/
	if (ch->direction() == hdr_cmn::UP) {
		
		dh = HDR_MAC_TDMA(p);  
		dst = ETHER_ADDR(dh->dh_da);


		
		/* frame utilization estimation */
		// bandest(ch->size_);

		// Since we can't really turn the radio off at lower level,
		// we just discard the packet when is not directed to us
		// or is broadcast
		if (dst != (u_int32_t) index_ && dst != MAC_BROADCAST) {
			Packet::free(p);
			return;
		}

		sendUp(p);
		return;
		
	}


	/* QS enabled MAC */
	if (enableQS_)
		checkQS(p);

	/* Packets coming down from ll layer (from ifq actually),
	   send them to phy layer. 
	   Now, it is in transmitting mode. */
	
	callback_ = h;	
	state(MAC_SEND);
	sendDown(p);

}

void MacTdmaDama::checkQS(Packet* p) {
	hdr_qs* qsh = hdr_qs::access(p);
	if (qsh->rate() > 0 && qsh->flag() == QS_REQUEST)
		requester_->totQSreq_ += 8 * hdr_qs::rate_to_Bps(qsh->rate());
}

void MacTdmaDama::sendUp(Packet* p) 
{

	hdr_qs* qsh = hdr_qs::access(p);
	/* Detect if there is any collision happened. should not happen...?*/
	if (rx_state_ == MAC_IDLE) {
		SET_RX_STATE(MAC_RECV);     // Change the state to recv.
		pktRx_ = p;                 // Save the packet for timer reference

		if (enableQS_ && qsh->rate() > 0 && qsh->flag() == QS_REQUEST) {
			SET_RX_STATE(MAC_IDLE);
			mhRxPkt_.start(p, delayQS_);
			return;
		}

		/* Start the timer for receiving, will end when receiving finishes. */
		mhRxPkt_.start(p, 0);
	} else {
		/* Note: we don't take the channel status into account, 
		   as collision should not happen...
		*/
		printf("<%d>, receiving, but the channel is not idle....???\n", index_);
	}

}

/* Actually receive data packet when RxPktTimer times out. */
void MacTdmaDama::recvDATA(Packet *p){

	/*Adjust the MAC packet size: strip off the mac header.*/
	struct hdr_cmn *ch = HDR_CMN(p);
	// ch->size() -= ETHER_HDR_LEN;
	ch->num_forwards() += 1;

	/* Pass the packet up to the link-layer.*/
	uptarget_->recv(p, (Handler*) 0);
}


void MacTdmaDama::queue_block(Packet* p) 
{
int numburst;
int last_burst_size;
int start_seqno, end_seqno;
int psize;
Packet* txp;

	hdr_cmn *ch = HDR_CMN(p);
	hdr_cmn *ch_f;
	hdr_tdmadama *mach_f;

	psize = ch->size() + MPE_HDR_LEN;
	printf("fragsize_ %d\n",fragsize_);

	last_burst_size = psize % fragsize_;
	numburst = (last_burst_size) ? psize / fragsize_ + 1 : psize / fragsize_;

	start_seqno = seqno_ + 1;
	end_seqno = seqno_ + numburst;

	for(int i = 0; i < numburst; i++ ) {
		txp = p->copy();

		ch_f = HDR_CMN(txp);
		mach_f = HDR_DAMA(txp);

		mach_f->psize() = ch->size();
		mach_f->bopno() = start_seqno;
		mach_f->eopno() = end_seqno;
		mach_f->numpk() = numburst;
	
		mach_f->ackrq() = 0;

		ch_f->size() = fragsize_;

		if ( i == numburst - 1 && last_burst_size )
			ch_f->size() = last_burst_size;

		mac_queue_->enque(txp);
		printf("size of the queue: %d\n",mac_queue_->length());
	}

}


/* Send packet down to the physical layer. 
   Need to calculate a certain time slot for transmission. */
void MacTdmaDama::sendDown(Packet* p) {
	u_int32_t dst, src, size;

	int send_byte;
	double send_time;

	struct hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac_tdma* dh = HDR_MAC_TDMA(p);

	dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
	dh->dh_fc.fc_type       = MAC_Type_Data;
	dh->dh_fc.fc_subtype    = MAC_Subtype_Data;

	dh->dh_fc.fc_to_ds      = 0;
	dh->dh_fc.fc_from_ds    = 0;
	dh->dh_fc.fc_more_frag  = 0;
	dh->dh_fc.fc_retry      = 0;
	dh->dh_fc.fc_pwr_mgt    = 0;
	dh->dh_fc.fc_more_data  = 0;
	dh->dh_fc.fc_wep        = 0;
	dh->dh_fc.fc_order      = 0;

	if((u_int32_t)ETHER_ADDR(dh->dh_da) != MAC_BROADCAST)
		dh->dh_duration = DATA_DURATION;
	else
		dh->dh_duration = 0;

	dst = ETHER_ADDR(dh->dh_da);
	src = ETHER_ADDR(dh->dh_sa);
	size = ch->size();

	if (pktTx_ != NULL)
		printf("<%d> warning: buffer is NOT empty\n",index_);


	/* buffer the packet to be sent. */
	// queue_block(p);
	pktTx_ = p;

	/* initialize the residue to whole packet length */
	residue_ = (HDR_CMN(pktTx_))->size();

	send_time = NOW - slot_start_time_;
	send_byte = slot_packet_len_-int(send_time*bandwidth_/8.0);


	if (slot_count_ >= 0 && ((send_time <= slot_time_ ) || hub_)
		&& (!aligned_ || residue_  <= send_byte ) ) {

		stat_loss();

		if (residue_ <= send_byte) {
			/* the packet can be sent immediately since is my turn,
			 * the MAC is idle and enough space is in slot*/
#if DEBUG_MAC
			trace_event("s-pack",residue_);
#endif

			/* is fragment lost ?? */

			send();
			residue_ = 0;
			return;
		}
#if DEBUG_MAC
		trace_event("s-frag",send_byte);
#endif

		residue_ -= send_byte;
	}
}

/* Actually send the packet. */
void MacTdmaDama::send() 
{
	u_int32_t dst, src, size;
	struct hdr_cmn* ch;
	struct hdr_mac_tdma* dh;
	double stime;

	/* Check if there is any packet buffered. */
	if (!pktTx_) {
		printf("<%d>, %f, no packet buffered.\n", index_, NOW);
		return;
	}

	/* Perform carrier sense...should not be collision...? */
	if(!is_idle()) {
		/* Note: we don't take the channel status into account, ie. no collision,
		   as collision should not happen...
		*/
		printf("<%d>, %f, transmitting, but the channel is not idle...???\n", index_, NOW);
		return;
	}

	ch = HDR_CMN(pktTx_);
	dh = HDR_MAC_TDMA(pktTx_);  

	dst = ETHER_ADDR(dh->dh_da);
	src = ETHER_ADDR(dh->dh_sa);
	size = ch->size();
	stime = TX_Time();
	
	/* Turn on the radio and transmit! */
	SET_TX_STATE(MAC_SEND);						     
	radioSwitch(ON);

	/* Start a timer that expires when the packet transmission is complete. */
	mhTxPkt_.start(pktTx_->copy(), stime);

	if (!packet_lost_) { 
		// printf("%.4lf MCS \n",NOW);
		downtarget_->recv(pktTx_, this);
	} else {
		// printf("%.4lf MDP \n",NOW);
		drop(pktTx_);
		packet_lost_ = 0;
	}

	pktTx_ = 0;

}

// Turn on / off the radio
void MacTdmaDama::radioSwitch(int i) 
{
	radio_active_ = i;
	//EnergyModel *em = netif_->node()->energy_model();
	if (i == ON) {
		//if (em && em->sleep())
		//em->set_node_sleep(0);
		//    printf("<%d>, %f, turn radio ON\n", index_, NOW); 
		Phy *p;
		p = netif_;
		((WirelessPhy *)p)->node_wakeup();
		return;
	}

	if (i == OFF) {
		//if (em && !em->sleep()) {
		//em->set_node_sleep(1);
		//    netif_->node()->set_node_state(INROUTE);
		Phy *p;
		p = netif_;
		((WirelessPhy *)p)->node_sleep();
		//    printf("<%d>, %f, turn radio OFF\n", index_, NOW);
		return;
	}
}


void MacTdmaDama::stat_loss() 
{
	if ( !flag_used_slot_ ) {
		flag_used_slot_ = 1;
		used_slots_++;
		if ( slot_lost_ )
			packet_lost_ = 1;
		else 
			successful_slots_++;
	}
}

/* This is the main TDMA - DAMA recursion */
void MacTdmaDama::slotHandler(Event *e) 
{
	slot_count_ = next_slot_;

	// Make the new preamble for next frame.
	if ( slot_count_ == max_slot_num_ ) {

		/* call the requester at the beginning of frame */
		re_schedule();

		/* call the allocator at the beginning of frame */
		re_allocate();

		slot_count_ = 0;
		
		/* if no packet is queued jump to the next frame 
		if (!pktTx_) {
			Scheduler::instance().schedule(&mhSlot_,(Packet *)e, 
				slot_time_*max_slot_num_);
			next_slot_ = max_slot_num_;
			return;
		}
		*/

	}

	// rescheduling for next slot 
	next_slot_ = slot_count_ + 1;
	while( !hub_ && 
		tdma_schedule_[next_slot_] != slot_num_ &&
		 next_slot_ < max_slot_num_ )
			next_slot_++;


	Scheduler::instance().schedule(&mhSlot_,(Packet *)e, 
		slot_time_*(next_slot_ - slot_count_));

	flag_used_slot_ = 0;


	// If it is the sending slot for me (or I am hub)
	if ( tdma_schedule_[slot_count_] == slot_num_ || hub_ ) {
	
		// printf("%lf %d %d\n",NOW,slot_num_,total_slots_);

		if ( em_  && em_->corrupt(pktTx_) ) {
			corrupted_slots_++;
			slot_lost_ = 1;
		} else
			slot_lost_ = 0;

		if (!pktTx_)
			return;

#if DEBUG_MAC
		trace_event("strslt",slot_num_);
#endif
		slot_start_time_ = NOW;
		stat_loss();

		if (residue_ <= slot_packet_len_) {
			
			/* the packet can be sent immediately since 
			 * this is my turn and we have enough space 
			 * in time slot */
	
#if DEBUG_MAC
			trace_event("s-pack",residue_);
#endif

			send();
			residue_ = 0;
			return;
		}
#if DEBUG_MAC
		trace_event("s-frag",slot_packet_len_);
#endif

		residue_ -= slot_packet_len_;
		return;
	}
}

void MacTdmaDama::recvHandler(Event *e) 
{

	u_int32_t dst, src; 
	Packet* pkt_=(Packet*)e;
	int size;
	struct hdr_cmn *ch = HDR_CMN(pkt_);
	struct hdr_mac_tdma *dh = HDR_MAC_TDMA(pkt_);

	/* Check if any collision happened while receiving. */
	if (rx_state_ == MAC_COLL) 
		ch->error() = 1;

	SET_RX_STATE(MAC_IDLE);
  
	/* check if this packet was unicast and not intended for me, drop it.*/   
	dst = ETHER_ADDR(dh->dh_da);
	src = ETHER_ADDR(dh->dh_sa);
	size = ch->size();

	// printf("<%d>, %f, recv a packet [from %d to %d], size = %d\n", index_, NOW, src, dst, size);

	// Turn the radio off after receiving the whole packet
	radioSwitch(OFF);

	/* Ordinary operations on the incoming packet */
	// Not a packet destinated to me.
	if ((dst != MAC_BROADCAST) && (dst != (u_int32_t)index_)) {
		drop(pkt_);
		return;
	}
  
	/* Now forward packet upwards. */
	recvDATA(pkt_);
}

/* After transmission a certain packet. Turn off the radio. */
void MacTdmaDama::sendHandler(Event *e) 
{
	
	/* Once transmission is complete, drop the packet. 
	   p  is just for schedule a event. */
	Packet::free((Packet *)e);
	
	/* unlock IFQ. */
	SET_TX_STATE(MAC_IDLE);
	
	if(callback_) {
		Handler *h = callback_;
		callback_ = 0;
		h->handle((Event*) 0);
	}
	
	/* Turn off the radio after sending the whole packet */
	radioSwitch(OFF);
}



/* Initialize Requester and Allocator*/ 
TdmaRequester::TdmaRequester():  qMon_(0)	
{
	reqerr_ = NULL;
	bind("request_",&request_);
	bind("totQSreq_",&totQSreq_);

}


TdmaAllocator::TdmaAllocator() : Connector()
{
	bind("delay_",&delay_);
	bind("hub_",&hub_);
	bind("active_node_",&active_node_);
	bind("max_slot_num_",&max_slot_num_);
	schedule_nxt_ = NULL;
}

AllocatorProportional::AllocatorProportional():	TdmaAllocator()
{
	deficit = 0;	
	slot_index = 0;
}

AllocatorFODA::AllocatorFODA():	TdmaAllocator()
{

	deficit = 0;	
	slot_index = 0;
	bind("def_slots_",&def_slots_);
	bind("ming_",&ming_);
	bind("maxg_",&maxg_);
	bind("k_frame_",&k_frame_);
	bind("xpercent_",&xpercent_);
	bind("ignore_loss_",&ignore_loss_);
	
	for(int i=0; i<MAC_MAX_TERM; i++ ) {
		def_all[i] = 0.0;
		req_register[i] = 0.0;
	}
	
	slotscycle = 0;
	service = 0;

}

/* ======================================================================
   TCL Hooks for the simulator
   ====================================================================== */

static class AllocatorProportionalClass : public TclClass {
public:
	AllocatorProportionalClass() : TclClass("Allocator/Proportional") {}
	TclObject* create(int, const char*const*) {
		return (new AllocatorProportional());
	}
} class_allocator_proportional;

static class AllocatorTclClass : public TclClass {
public:
	AllocatorTclClass() : TclClass("Allocator/Tcl") {}
	TclObject* create(int, const char*const*) {
		return (new AllocatorTcl());
	}
} class_allocator_tcl;

static class AllocatorFODAClass : public TclClass {
public:
	AllocatorFODAClass() : TclClass("Allocator/FODA") {}
	TclObject* create(int, const char*const*) {
		return (new AllocatorFODA());
	}
} class_allocator_foda;

static class RequesterConstantClass : public TclClass {
public:
	RequesterConstantClass() : TclClass("Requester/Constant") {}
	TclObject* create(int, const char*const*) {
		return (new RequesterConstant());
	}
} class_requester_constant;

static class RequesterRBDCClass : public TclClass {
public:
	RequesterRBDCClass() : TclClass("Requester/RBDC") {}
	TclObject* create(int, const char*const*) {
		return (new RequesterRBDC());
	}
} class_requester_rbdc;

static class RequesterVBDCClass : public TclClass {
public:
	RequesterVBDCClass() : TclClass("Requester/VBDC") {}
	TclObject* create(int, const char*const*) {
		return (new RequesterVBDC());
	}
} class_requester_vbdc;

static class RequesterRVBDCClass : public TclClass {
public:
	RequesterRVBDCClass() : TclClass("Requester/RVBDC") {}
	TclObject* create(int, const char*const*) {
		return (new RequesterRVBDC());
	}
} class_requester_rvbdc;

static class RequesterTclClass : public TclClass {
public:
	RequesterTclClass() : TclClass("Requester/Tcl") {}
	TclObject* create(int, const char*const*) {
		return (new RequesterTcl());
	}
} class_requester_tcl;


RequesterRVBDC::RequesterRVBDC() : TdmaRequester() {
	bind("alpha_",&alpha_);
	bind("k_",&k_);
	bind("max_rvbdc_",&max_rvbdc_);
	bind("min_rvbdc_",&min_rvbdc_);
        pin = 0;
	tin = 0;
}
	
/* Reinitialize Allocator and Requester objects after new nodes is created */
void TdmaAllocator::init() {
	
	requestv_ = mac_->requestv_;
	max_slot_num_ = mac_->max_slot_num_;
	tdma_schedule_nxt_ = mac_->tdma_schedule_nxt_;
	if (schedule_nxt_)
		delete [] schedule_nxt_;
	schedule_nxt_ = new int[max_slot_num_];
	active_node_ = mac_->active_node_;

	mac_->allocator_ = this;

}

void AllocatorProportional::init(){

	if (deficit) 
		delete[] deficit;
		
	active_node_ = mac_->active_node_;

	deficit = new double[active_node_];
	for(int i=0; i<active_node_; i++)
		deficit[i] = 0;	

	TdmaAllocator::init();
}

void AllocatorFODA::init(){

	if (deficit) 
		delete[] deficit;
		
	active_node_ = mac_->active_node_;

	deficit = new double[active_node_];
	for(int i=0; i<active_node_; i++)
		deficit[i] = 0;	

	TdmaAllocator::init();
}

void AllocatorTcl::init(){
	
	Tcl& tcl = Tcl::instance();

	active_node_ = mac_->active_node_;

	// initialize AllocatorTcl instvars
	tcl.evalf("%s initvar",name());

	TdmaAllocator::init();
}

void TdmaRequester::init() {
	requestv_ = mac_->requestv_;
	slot_num_ = mac_->slot_num_;
	mac_->requester_ = this;
}

void RequesterTcl::init() {
	
	Tcl& tcl = Tcl::instance();

	// initialize RequesterTcl instvars
	tcl.evalf("%s initvar",name());

	TdmaRequester::init();
}


/* ======================================================================
   Command Interface with the interpreter 
   ====================================================================== */
int AllocatorProportional::command(int argc, const char*const* argv)
{

	if (argc == 3) {
		if (strcmp(argv[1], "attach") == 0) {
			mac_ = (MacTdmaDama*) TclObject::lookup(argv[2]);
			if(mac_ == 0)
				return TCL_ERROR;
			init();
			return TCL_OK;
		}
	}
	return TdmaAllocator::command(argc, argv);
}

int AllocatorFODA::command(int argc, const char*const* argv)
{

	if (argc == 3) {
		if (strcmp(argv[1], "attach") == 0) {
			mac_ = (MacTdmaDama*) TclObject::lookup(argv[2]);
			if(mac_ == 0)
				return TCL_ERROR;
			init();
			return TCL_OK;
		}
	}
	if (argc == 4) {
		if (strcmp(argv[1], "assign") == 0) {
			int term = atoi(argv[2]);
			double ass = atof(argv[3]);
			def_all[term] = ass;
			return TCL_OK;
		}
	}
	return TdmaAllocator::command(argc, argv);
}

int AllocatorTcl::command(int argc, const char*const* argv)
{

	if (argc == 3) {
		if (strcmp(argv[1], "attach") == 0) {
			mac_ = (MacTdmaDama*) TclObject::lookup(argv[2]);
			if(mac_ == 0)
				return TCL_ERROR;
			init();
			return TCL_OK;
		}
	}
	return TdmaAllocator::command(argc, argv);
}


int TdmaRequester::command(int argc, const char*const* argv){
	
	if (argc == 3) {
		if (strcmp(argv[1], "attach") == 0) {
			mac_ = (MacTdmaDama*) TclObject::lookup(argv[2]);
			if(mac_ == 0)
				return TCL_ERROR;
			init();	
			return TCL_OK;
		}
		if (strcmp(argv[1], "setifq") == 0) {
			ifq_ = (Queue *)TclObject::lookup(argv[2]);
			if (!ifq_)
				return TCL_ERROR;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "setmon") == 0) {
			qMon_ = (QueueMonitor *)TclObject::lookup(argv[2]);
			if (!qMon_)
				return TCL_ERROR;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "reqerr") == 0) {
			reqerr_ = (ErrorModel *)TclObject::lookup(argv[2]);
			if (!reqerr_)
				return TCL_ERROR;
			return (TCL_OK);
		}
	}
	return TclObject::command(argc,argv);
}

int RequesterConstant::command(int argc, const char*const* argv)
{
	return TdmaRequester::command(argc, argv);
}

int RequesterRBDC::command(int argc, const char*const* argv)
{
	return TdmaRequester::command(argc, argv);
}

int RequesterVBDC::command(int argc, const char*const* argv)
{
	return TdmaRequester::command(argc, argv);
}

int RequesterRVBDC::command(int argc, const char*const* argv)
{
	return TdmaRequester::command(argc, argv);
}

int RequesterTcl::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "attach") == 0) {
			mac_ = (MacTdmaDama*) TclObject::lookup(argv[2]);
			if(mac_ == 0)
				return TCL_ERROR;
			init();	
			return TCL_OK;
		}
	}
	return TdmaRequester::command(argc, argv);
}


/* ======================================================================
   Debugging Routines
   ====================================================================== */

void TdmaAllocator::printrequests(){

	for(int i=0; i<active_node_ ; i++ )
	   printf("[%lf]",requestv_[i]);
	printf("\n");

}

void TdmaAllocator::printrequest_nxt(){
	for(int i=0; i<max_slot_num_ ; i++ )
  		if (schedule_nxt_[i] == SLOT_UNALLOC)
			printf("[N]");
		else
			printf("[%d]", schedule_nxt_[i]);
	printf("\n");
}


void TdmaRequester::req_error_model() {

	if (!reqerr_)
		return;

	if (reqerr_->corrupt(NULL)) {
 		requestv_[slot_num_] = REQ_LOST;
		// printf("lost request\n");
	}

}

/* ======================================================================
   TDMA routines to calculate requests 
   ====================================================================== */
void RequesterConstant::dorequest(){

	/* overwrite previous request */
	requestv_[slot_num_] = request_;

	req_error_model();

}

void RequesterRBDC::dorequest(){

	rate = (double(qMon_->barrivals() - pin))/
                (mac_->slot_packet_len_*mac_->num_frame_);
	
	/* overwrite previous request */
	requestv_[slot_num_] = rate;

	pin = qMon_->barrivals();
	tin = NOW;

	req_error_model();

}

void RequesterVBDC::dorequest(){

	buffer = qMon_->barrivals() - qMon_->bdepartures() - qMon_->bdrops();
	/* overwrite previous request */
	printf("%lf %d %d\n",NOW, slot_num_, 
	  buffer/(mac_->slot_packet_len_*mac_->num_frame_));
	requestv_[slot_num_] = buffer/(mac_->slot_packet_len_*mac_->num_frame_)+1;

	req_error_model();
}

void RequesterRVBDC::dorequest(){
double buffer_comp;
double rate_comp;

	rate = qMon_->barrivals() - pin;
	buffer = qMon_->barrivals() - qMon_->bdepartures() - qMon_->bdrops();

	pin = qMon_->barrivals();
	tin = NOW;

	buffer_comp = buffer/(mac_->slot_packet_len_*mac_->num_frame_);
	rate_comp = rate/(mac_->slot_packet_len_*mac_->num_frame_);

	/* override previous request */
	requestv_[slot_num_] =  alpha_ * rate_comp + (1 - alpha_) 
			* buffer_comp + min_rvbdc_;
	
#if 0
	if (requestv_[slot_num_] > max_rvbdc_ )
		requestv_[slot_num_] =  max_rvbdc_;

	if (requestv_[slot_num_] < min_rvbdc_ )
		requestv_[slot_num_] = min_rvbdc_;

	printf("%d %.3lf %.3lf %.3lf %.3lf\n",
	slot_num_,NOW,buffer_comp,rate_comp,requestv_[slot_num_]);
#endif

	req_error_model();
}

void RequesterTcl::dorequest(){

	char wrk[1024];
	Tcl& tcl = Tcl::instance();
	
	// call the requester tcl
	sprintf(wrk,"%s dorequest",name());
	tcl.evalc(wrk);

	/* overwrite previous request */
	requestv_[slot_num_] = atof(tcl.result());
	
	req_error_model();
}

/* ======================================================================
   TDMA Allocation Functions
   ====================================================================== */
void AllocationDelay::handle(Event* e) {

        if (status_ != TIMER_PENDING)   // sanity check
                abort();
        status_ = TIMER_HANDLING;
        expire(e);
        // if it wasn't rescheduled, it's done
        if (status_ == TIMER_HANDLING)
                status_ = TIMER_IDLE;

	delete this;
}


void TdmaAllocator::allocation(){

	/* create new timer */
	copy_ = new AllocationDelay(this);

	/* store schedule_nxt to local memory */
	copy_->schedule_ = new int[max_slot_num_];

	for(int i=0; i<max_slot_num_; i++)
		copy_->schedule_[i] = schedule_nxt_[i];
	
	/* schedule the allocation */
	copy_->sched(delay_);
        
	mac_->trace_event("reques",0);
	mac_->trace_event("nxt",0);

}

void TdmaAllocator::schedule_copy(AllocationDelay* c){
	
	/* copy the sent schedule to local MAC */
	for(int i=0; i<max_slot_num_; i++) 
		tdma_schedule_nxt_[i] = c->schedule_[i];

	mac_->trace_event("cur",0);

}

/* Implements the Proportional Allocation */
void AllocatorProportional::allocation() 
{
	double totreq = 0;		// sum of all requests
	int slot = 0;	
	int ns = 0;
	int si = slot_index;

	// find the sum of all requests
	for(int i=0; i<active_node_; i++)  {
	   totreq += requestv_[i];

	}

	// add the previous deficit to current request
	if ( totreq > 0 )
	   for(int i=0; i<active_node_; i++) 
	      if (i!=slot_index)
		 deficit[i] += max_slot_num_*requestv_[i]/totreq;
	  
	while( slot<max_slot_num_ && ns<active_node_ ) {
	
	    while( ns<active_node_ && deficit[slot_index]<=EPS ) {
		slot_index = (slot_index + 1) % active_node_;
		ns++;
	    }

	    while( slot<max_slot_num_ && deficit[slot_index]>EPS ) {
	        deficit[slot_index]--;
	        schedule_nxt_[slot] = slot_index;
	        slot++;
	    }
	}

	deficit[si] += max_slot_num_*requestv_[si]/totreq;

	// the rest is not allocated
	if (slot<max_slot_num_)
	   for(int i=slot; i<max_slot_num_; i++)
	      if (deficit[si]>EPS) {
		 deficit[si]--;
		 schedule_nxt_[i] = si; 
	      } else 
	         schedule_nxt_[i] = SLOT_UNALLOC;

	/* operation common to all Allocators */
	TdmaAllocator::allocation();
}

void AllocatorFODA::allocation() 
{
	int slot = -1;                   // slot counter	

	/* updating requests */
	for(int i=0; i<active_node_; i++) { 

	     if ( !(requestv_[i] < 0) || ignore_loss_ )
	     	req_register[i] = requestv_[i]; 

	     /* a new request was received, 
	      * update the register */
	     if ( req_register[i] > maxg_/xpercent_ )
	 	req_register[i] = maxg_/xpercent_;

 	     if ( req_register[i] < ming_/xpercent_ )
 	        req_register[i] = ming_/xpercent_;
	
	}

	// printf("%.4lf %.4lf %.4lf\n",NOW,requestv_[1],req_register[1]);
	
	/* emulation of load (station zero represent the others) */
	req_register[0] = (k_frame_ * max_slot_num_)/xpercent_ - req_register[1];

	/* slot assignment in this frame */
	while( slot < max_slot_num_ ) {
		
		slot++;	

		if ( service <= 0 ) {
			slot_index = (slot_index + 1) % active_node_;
			service = (int)(xpercent_ * req_register[slot_index] + 0.5);
			req_register[slot_index] -= service;
		}

		schedule_nxt_[slot] = slot_index;
		service--;
	}

	/* operation common to all Allocators */
	TdmaAllocator::allocation();
}

/* Implements the Tcl Embedded allocation function */ 
void AllocatorTcl::allocation() 
{
	Tcl & tcl = Tcl::instance();

	char wrk[1024];
	
	// initialize the array
	for(int i=0; i<active_node_; i++)
		tcl.evalf("%s set requestv_(%d) %lf",name(),i,requestv_[i]);

	// call Tcl scheduling function
	tcl.evalf("%s allocation", name());

	for(int i=0; i<max_slot_num_; i++){
		sprintf(wrk,"%s set schedule_(%d)",name(),i);
		tcl.evalc(wrk);
		schedule_nxt_[i] = atoi(tcl.result());
	}
	
	// operation common to all Allocators 
	TdmaAllocator::allocation();
}


