/*
 * Copyright (C) 1999 by the University of Southern California
 * $Id: mac-tdma.h,v 1.5 2005/08/25 18:58:07 johnh Exp $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 * The copyright of this module includes the following
 * linking-with-specific-other-licenses addition:
 *
 * In addition, as a special exception, the copyright holders of
 * this module give you permission to combine (via static or
 * dynamic linking) this module with free software programs or
 * libraries that are released under the GNU LGPL and with code
 * included in the standard release of ns-2 under the Apache 2.0
 * license or under otherwise-compatible licenses with advertising
 * requirements (or modified versions of such code, with unchanged
 * license).  You may copy and distribute such a system following the
 * terms of the GNU GPL for this module and the licenses of the
 * other code concerned, provided that you include the source code of
 * that other code when and as the GNU GPL requires distribution of
 * source code.
 *
 * Note that people who make modified versions of this module
 * are not obligated to grant this special exception for their
 * modified versions; it is their choice whether to do so.  The GNU
 * General Public License gives permission to release a modified
 * version without this exception; this exception also makes it
 * possible to release a modified version which carries forward this
 * exception.
 *
 */

#ifndef ns_mac_tdmadama_h
#define ns_mac_tdmadama_h

// #define DEBUG
// #include <debug.h>

#include "marshall.h"
#include "queue-monitor.h"

#include <delay.h>
#include <connector.h>
#include <packet.h>
#include <random.h>
#include <arp.h>
#include <ll.h>
#include <mac.h>
#include <hdr_qs.h>
#include <errmodel.h>

#define GET_ETHER_TYPE(x)		GET2BYTE((x))
#define SET_ETHER_TYPE(x,y)            {u_int16_t t = (y); STORE2BYTE(x,&t);}

/* We are still using these specs for phy layer---same as 802.11. */
/*
 * IEEE 802.11 Spec, section 15.3.2
 *	- default values for the DSSS PHY MIB
 */
#define DSSS_CWMin			31
#define DSSS_CWMax			1023
#define DSSS_SlotTime			0.000020	// 20us
#define	DSSS_CCATime			0.000015	// 15us
#define DSSS_RxTxTurnaroundTime		0.000005	// 5us
#define DSSS_SIFSTime			0.000010	// 10us
#define DSSS_PreambleLength		144		// 144 bits
#define DSSS_PLCPHeaderLength		48		// 48 bits

#define EPS                             0.0000001

class PHY_MIB {
public:
	u_int32_t	CWMin;
	u_int32_t	CWMax;
	double		SlotTime;
	double		CCATime;
	double		RxTxTurnaroundTime;
	double		SIFSTime;
	u_int32_t	PreambleLength;
	u_int32_t	PLCPHeaderLength;
};

/* ======================================================================
   Frame Formats
   ====================================================================== */

#define	MAC_ProtocolVersion	0x00
#define MAC_Type_Data		0x02
#define MAC_Subtype_Data	0x00

// Max data length allowed in one slot (byte)
#define MAC_TDMA_MAX_DATA_LEN 1500        

// Max number of terminals in FODA/IBEA
#define MAC_MAX_TERM         1000

// How many time slots in one frame.
#define MAC_TDMA_SLOT_NUM       32           

// The mode for MacTdmaDama layer's defer timers. */
#define SLOT_SCHE               0
#define SLOT_SEND               1
#define SLOT_RECV               2
#define SLOT_BCAST              3

// Indicate if there is a packet needed to be sent out.
#define NOTHING_TO_SEND         -2
// Indicate if this is the very first time the simulation runs.
#define FIRST_ROUND             -1

// Turn radio on /off
#define ON                       1
#define OFF                      0

/* Quoted from MAC-802.11. */
#define DATA_DURATION           5

// Indicate that the slot is not allocated 
#define SLOT_UNALLOC             -2
#define SLOT_NOREQ               0
#define REQ_LOST                 -1

/* MAC header format */
#define MPE_HDR_LEN    16
#define ULE_HDR_LEN    4
#define HDR_DAMA(p)    (hdr_tdmadama::access(p))
#define SEQNO(p)       HDR_DAMA((p)->pkt)->seqno()
#define SN(p)          HDR_DAMA(p)->seqno()
#define BOPNO(p)       HDR_DAMA(p)->bopno()
#define EOPNO(p)       HDR_DAMA(p)->eopno()
#define NUMPK(p)       HDR_DAMA(p)->numpk()

/* TDMA DAMA header */
struct hdr_tdmadama {
	LLFrameType lltype_;		// link-layer frame type
	int seqno_;			// sequence number
	int ackno_;			// acknowledgement number
	int bopno_;			// begin of packet seqno
	int eopno_;			// end of packet seqno
	int psize_;			// size of packet
	int ackrq_;			// arq requested_ 
	int numpk_;			// number of packets (for PLFEC)
	double sendtime_;		// time the packet is sent
	double spacing_;		// packet has been delayed

	static int offset_;
	inline int& offset() { return offset_; }
	static hdr_tdmadama* access(const Packet* p) {
		return (hdr_tdmadama*) p->access(offset_);
	}

	inline LLFrameType& lltype() { return lltype_; }
	inline int& seqno() { return seqno_; }
	inline int& ackno() { return ackno_; }
	inline int& bopno() { return bopno_; }
	inline int& eopno() { return eopno_; }
	inline int& psize() { return psize_; }
	inline int& ackrq() { return ackrq_; }
	inline int& numpk() { return numpk_; }
	inline double& sendtime() { return sendtime_; }
	inline double& spacing() { return spacing_; }
};


/* We are using same header structure as 802.11 currently */
struct frame_control {
	u_char		fc_subtype		: 4;
	u_char		fc_type			: 2;
	u_char		fc_protocol_version	: 2;

	u_char		fc_order		: 1;
	u_char		fc_wep			: 1;
	u_char		fc_more_data		: 1;
	u_char		fc_pwr_mgt		: 1;
	u_char		fc_retry		: 1;
	u_char		fc_more_frag		: 1;
	u_char		fc_from_ds		: 1;
	u_char		fc_to_ds		: 1;
};

struct hdr_mac_tdma {
	struct frame_control	dh_fc;
	u_int16_t		dh_duration;
	u_char			dh_da[ETHER_ADDR_LEN];
	u_char			dh_sa[ETHER_ADDR_LEN];
	u_char			dh_bssid[ETHER_ADDR_LEN];
	u_int16_t		dh_scontrol;
	u_char			dh_body[1]; // XXX Non-ANSI
};

#define DATA_Time(len)	(8 * (len) / bandwidth_)

/* ======================================================================
   The following destination class is used for duplicate detection.
   ====================================================================== */

class MacTdmaDama;


/* Make a request to allocator for each Tdma node */
class TdmaRequester: public TclObject {

public:
	TdmaRequester();
	virtual void dorequest() {};

	int command(int argc, const char*const* argv);
	
	// initialize dynamic variables
	void init();

	/* pointer to Queue */
  	Queue* ifq_;
	QueueMonitor* qMon_;
	
	/* QS variables */
	double totQSreq_;

protected:
	MacTdmaDama* mac_;
	int slot_num_;
	double *requestv_;

      	/* the request to submit */
	double request_;
	
	virtual void req_error_model();
	ErrorModel* reqerr_;
	
};


class RequesterConstant: public TdmaRequester {

public:
	RequesterConstant(): TdmaRequester() { };
	int command(int argc, const char*const* argv);
	
	virtual void dorequest();

};

class RequesterRBDC: public TdmaRequester {

public:
	RequesterRBDC(): TdmaRequester() { };
	int command(int argc, const char*const* argv);
	
	virtual void dorequest();
	
private:
	double rate;
        int pin;
	double tin;
};

class RequesterVBDC: public TdmaRequester {

public:
	RequesterVBDC(): TdmaRequester() { };
	int command(int argc, const char*const* argv);
	
	virtual void dorequest();

private:
       int buffer;
};

class RequesterRVBDC: public TdmaRequester {

public:
	RequesterRVBDC();
	int command(int argc, const char*const* argv);
	
	virtual void dorequest();

private:
        int buffer;
        double rate;
        double alpha_;
        double k_;
	double max_rvbdc_;
	double min_rvbdc_;

        int pin;
	double tin;

};

class RequesterTcl: public TdmaRequester {

public:
	RequesterTcl(): TdmaRequester() { };
	int command(int argc, const char*const* argv);
	
	virtual void dorequest();

	void init();
};

class AllocationDelay;

/* Allocator classes */
class TdmaAllocator: public Connector {
friend class AllocationDelay;

public:
	TdmaAllocator();
	int command(int argc, const char*const* argv){ 
		return Connector::command(argc,argv); 
	};

	void virtual allocation();
	int active_node_;
	int max_slot_num_;
	int hub_;           // allocator logical addr
	
	void virtual init();
	int *schedule_nxt_;

protected:
	MacTdmaDama* mac_;
	
	/* request vector */
	double *requestv_;
	int *tdma_schedule_nxt_;
	void printrequests();
	void printrequest_nxt();

	/* allocation delay */
	double delay_;
	double jitter_;

	AllocationDelay* copy_;
	void schedule_copy(AllocationDelay *);

	BaseTrace* bt;	
};

/* implements the timer that copy next allocation */
class AllocationDelay: public TimerHandler {
public:
	AllocationDelay(TdmaAllocator* a): TimerHandler() {
		allocator_ = a;
	}
	int * schedule_;
	
	virtual void expire(Event *e){
		allocator_->schedule_copy(this);
	}

	virtual void handle(Event *);
	
	~AllocationDelay() {
		delete [] schedule_;
	}

protected:
	TdmaAllocator* allocator_;
};

class AllocatorProportional: public TdmaAllocator {

public:
	AllocatorProportional();
	int command(int argc, const char*const* argv);
	virtual void allocation();
	void virtual init();
	
protected:
  	double* deficit;	          // deficit for allocator
	int slot_index;
	double creq;
};

class AllocatorFODA: public TdmaAllocator {

public:
	AllocatorFODA();
	int command(int argc, const char*const* argv);
	virtual void allocation();
	void virtual init();
	
protected:
  	double* deficit;	          // deficit for allocator
	int slot_index;
	int def_slots_;
	int ming_;
	int maxg_;
	double creq;

	/* default allocations */	
	double def_all[MAC_MAX_TERM];
	double req_register[MAC_MAX_TERM];
	double xpercent_;
	int ignore_loss_;

	/* current amount of slot to serve */
	int service;

	/* emulation purposes */
	double k_frame_;		// estimation of allocation cycle length
	int slotscycle;			// number of slots in an allocation cycle

};

class AllocatorTcl: public TdmaAllocator {

public:
	AllocatorTcl(): TdmaAllocator() {} ;
	int command(int argc, const char*const* argv);
	virtual void allocation();
	void virtual init();
protected:
	int slot_index;
	
};

class MacTdmaDamaTimer : public Handler {
public:
	MacTdmaDamaTimer(MacTdmaDama* m, double s = 0) : mac(m) {
		busy_ = paused_ = 0; stime = rtime = 0.0; slottime_ = s;
	}

	virtual void handle(Event *e) = 0;

	virtual void start(Packet *p, double time);
	virtual void stop(Packet *p);
	virtual void pause(void) { assert(0); }
	virtual void resume(void) { assert(0); }

	inline int busy(void) { return busy_; }
	inline int paused(void) { return paused_; }
	inline double slottime(void) { return slottime_; }
	inline double expire(void) {
		return ((stime + rtime) - Scheduler::instance().clock());
	}


protected:
	MacTdmaDama 	*mac;
	int		busy_;
	int		paused_;
	Event		intr;
	double		stime;	// start time
	double		rtime;	// remaining time
	double		slottime_;
};

/* Timers to schedule transmitting and receiving. */
class SlotTdmaDamaTimer : public MacTdmaDamaTimer {
public:
	SlotTdmaDamaTimer(MacTdmaDama *m) : MacTdmaDamaTimer(m) {}
	void	handle(Event *e);
};

/* Timers to control packet sending and receiving time. */
class RxPktTdmaDamaTimer : public MacTdmaDamaTimer {
public:
	RxPktTdmaDamaTimer(MacTdmaDama *m) : MacTdmaDamaTimer(m) {}
	void	handle(Event *e);
};

class TxPktTdmaDamaTimer : public MacTdmaDamaTimer {
public:
	TxPktTdmaDamaTimer(MacTdmaDama *m) : MacTdmaDamaTimer(m) {}
	void	handle(Event *e);
};

/* TDMA Mac layer. */
class MacTdmaDama : public Mac {
  friend class SlotTdmaDamaTimer;
  friend class TxPktTdmaDamaTimer;
  friend class RxPktTdmaDamaTimer;
  friend class TdmaAllocator;
  
 public:
  MacTdmaDama(PHY_MIB* p);
  void		recv(Packet *p, Handler *h);
  inline int	hdr_dst(char* hdr, int dst = -2);
  inline int	hdr_src(char* hdr, int src = -2);
  inline int	hdr_type(char* hdr, u_int16_t type = 0);
  
  /* Timer handler */
  void slotHandler(Event *e);
  void recvHandler(Event *e);
  void sendHandler(Event *e);
  
  static double *requestv_;             // The vector of requests done by stations
  
  /* Requester and Allocator objects for slot scheduling */
  static TdmaAllocator* allocator_;
  TdmaRequester* requester_;

  /* Data structure for tdma scheduling. */
  static int active_node_;            // How many nodes needs to be descheduled

  static int *tdma_schedule_nxt_;     // The next frame to allocate
  
  /* TDMA scheduling state. 
     Currently, we only use a centralized simplified way to do 
     scheduling. Will work on the algorithm later.*/
  // The max num of slot within one frame.
  static int max_slot_num_;
  int slot_num_;                      // The slot number it's allocated.
  int fragsize_;
  int seqno_; 

  /* frame utilization*/
  void bandest(int);
  double utilization();
  double peak_utilization();
  double abw_;               // estimated available bandwidth
  double int_begin;
  double util_weight_;
  int algorithm_;
  int fixed_frame_;
  int slots1_;
  int offset_;
  int m_frame_;
  int m_Fcount;

  /* QS variables */
  int enableQS_;
  double delayQS_;
  void checkQS(Packet*);

  /* MAC Trace support */
  EventTrace* et_;
  virtual void trace_event(char* event,int val);
  void stat_loss();
  
 // Both the slot length and max slot num (max node num) can be configged.
  int			slot_packet_len_;
 
 // The number of frame after which a request is done
  static int num_frame_;

  /* manage packet error models */
  ErrorModel* em_;
  ErrorModel* btp_; 
 
 protected:
  PHY_MIB		*phymib_;
  
 private:
  int command(int argc, const char*const* argv);

  // dump only allocation
  int dump_alloc_;
  
  // Do slot scheduling for the active nodes within one cluster.
  void re_allocate();
  void re_schedule();
  void makePreamble();
  void radioSwitch(int i);

  /* Packet Transmission Functions.*/
  void    sendUp(Packet* p);
  void    sendDown(Packet* p);
  void    queue_block(Packet* p);
  Queue*  mac_queue_;

  /* Actually receive data packet when rxTimer times out. */
  void recvDATA(Packet *p);
  /* Actually send the packet buffered. */
  void send();

  inline int	is_idle(void);
  
  /* Debugging Functions.*/
  void		trace_pkt(Packet *p);
  void		dump(char* fname);
  
  void mac_log(Packet *p) {
    logtarget_->recv(p, (Handler*) 0);
  }
  
  inline double TX_Time() {
    double t = DATA_Time(residue_);
    
    return t;
  }
  
  inline u_int16_t usec(double t) {
    u_int16_t us = (u_int16_t)ceil(t *= 1e6);
    return us;
  };

  /* Timers */
  SlotTdmaDamaTimer mhSlot_;
  TxPktTdmaDamaTimer mhTxPkt_;
  RxPktTdmaDamaTimer mhRxPkt_;


   /* Internal MAC state */
  MacState	rx_state_;	// incoming state (MAC_RECV or MAC_IDLE)
  MacState	tx_state_;	// outgoing state
  
  /* The indicator of the radio. */
  int radio_active_;

  NsObject*	logtarget_;
  
  int frame_;
  
  static double slot_time_;   // The time duration for each slot.
  static double start_time_;  // The start time for whole TDMA scheduling
  double slot_start_time_;    // initial slot time 
  
  int tdma_schedule_[MAC_MAX_TERM];
	   
  int slot_count_;           // When slot_count_ = active_nodes_, a new preamble is needed
  int next_slot_;            // the next scheduled slot 
 
  int residue_;               // The remaining part of packet to send
  int packet_lost_;
  int slot_lost_;
  int aligned_;               // packet aligned with slot
  int hub_;
  static int num_requests_;   // number of request received during active frame
  
  static int tdma_ps_;        // How many packets has been sent out?
  static int tdma_pr_;
  
  /* statistics */
  int corrupted_slots_;
  int used_slots_;
  int successful_slots_;
  int total_slots_;
  int flag_used_slot_;

  void watch_alloc();

};

#endif /* __mac_tdmadama_h__ */

